--[[ This file stores the basic functions for the casthistory.
Instead like in the first version of the cast history, this time I completely detached the history from the castbar, this way it is much more flexible, reliable and much less
vulnerable to errors. ]]--
local defaultxOffsetModifier = 3;

function ArenaLive_CastHistory_Initialise(self, addonName, frameType, numIcons, iconTemplate, moveDirection)
	
	-- Create a shortcut for the parent frame.
	local parent = self:GetParent();
	parent.castHistory = self;

	local prefix = self:GetName();
	self.addonName = addonName;
	self.frameType = frameType;
	self.numIcons = numIcons;
	self.iconTemplate = iconTemplate;
	self.moveDirection = moveDirection;
	
	-- We create a defined number of icons right at the beginning, most of the time it will be the same number as the maximal shown history icon number.
	for i = 1, numIcons do

		local iconName = "Icon"..i;
		local iconShortCut = "icon"..i;
		ArenaLive_CastHistoryIcon_Create(self, iconName, iconShortCut, moveDirection, iconTemplate);
		
	end	
	
	self:RegisterEvent("UNIT_SPELLCAST_START");
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
	self:RegisterEvent("UNIT_SPELLCAST_STOP");
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START");
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP");

end



function ArenaLive_CastHistory_UpdateUnit (self, unit)
	
	if ( not unit ) then
		return;
	end

	self.unit = unit;
	self.guid = UnitGUID(unit); -- Since Combatlog doesn't work with unitIDs, we need to save guid too.
	ArenaLive_CastHistory_ResetAllIcons(self);
	self:SetScript("OnEvent",  ArenaLive_CastHistory_OnEvent);
	self:Show();

end



function ArenaLive_CastHistory_ResetFrame(self)
	ArenaLive_SystemMessage("Reset castHistory: "..self:GetName(), true);
	self.unit = nil;
	self.guid = nil;
	self.isCasting = nil;
	self.lineID = nil;
	self.isChanneling = nil;
	self:SetScript("OnEvent",  nil);
	ArenaLive_CastHistory_ResetAllIcons(self);
	self:Hide();

end



function ArenaLive_CastHistory_OnEvent(self, event, ... )
	
	local unit, spell, rank, lineID, spellID  = ...;
	
	if ( not self.unit or ( unit ~= self.unit ) or not self:IsVisible() ) then
		return;
	end
	
	if ( event == "UNIT_SPELLCAST_START" ) then
		self.isCasting = true;
		self.lineID = lineID;
		ArenaLive_CastHistory_RotateIcons(self, "hardCast", spellID, lineID);
		
	elseif ( event == "UNIT_SPELLCAST_SUCCEEDED" ) then
		if ( self.isCasting and lineID == self.lineID ) then
			self.isCasting = nil;
			self.lineID = nil;
		elseif ( ( not self.isCasting ) and ( not self.isChanneling ) and (not ARENALIVE_SPELLS["filteredSpells"][spellID] ) ) then
			
			-- We check if we are channeling, because "UNIT_SPELLCAST_SUCCEEDED" normally fires for every tick of a channeled spell.
			ArenaLive_CastHistory_RotateIcons(self, "instant", spellID, lineID);
		end
	
	elseif ( ( event == "UNIT_SPELLCAST_STOP" ) and ( lineID == self.lineID ) ) then
			self.isCasting = nil;
			self.lineID = nil;
	elseif ( event == "UNIT_SPELLCAST_CHANNEL_START" ) then
		self.isChanneling = true;
		ArenaLive_CastHistory_RotateIcons(self, "channel", spellID, lineID);
	elseif ( event == "UNIT_SPELLCAST_CHANNEL_STOP" ) then
		self.isChanneling = false;
	end
end



function ArenaLive_CastHistory_RotateIcons(castHistory, castType, spellID, lineID)

	local name, rank, icon, powerCost, isFunnel, powerType, castingTime, minRange, maxRange = GetSpellInfo(spellID);
	local iconFound = nil;
	local historyIcon;
	local dontMove = true;
	
	-- It is possible, that there is no visible frame on the initial position, because we fadeout failed casts all the time.
	-- If this is the case, we don't need to move anything, because we already have enough space.	
	for i = 1, castHistory.numIcons do
		historyIcon = castHistory["icon"..i];
		
		local point, relativeTo, relativePoint, xOffset, yOffset =  historyIcon:GetPoint();
		if ( ( xOffset == 0 and yOffset == 0 ) and ( historyIcon.isActive ) and ( historyIcon:IsShown() ) ) then
			dontMove = nil;
			break;
		end
			
	end

	historyIcon = nil;
	
	for i = 1, castHistory.numIcons do
	
		historyIcon = castHistory["icon"..i];
		
		if ( not historyIcon.isActive ) then
	
			if ( not iconFound ) then
				
				historyIcon.spellID = spellID;
				
				-- PvP-Trinekt spell has a different icon than the trinkets. So I replace it here with the according one.
				if ( spellID == ARENALIVE_SPELLS["trinket"] ) then
					local _, faction = UnitFactionGroup(castHistory.unit);
					
					if ( faction == "Alliance" ) then
						icon = ("Interface\\ICONS\\INV_Jewelry_TrinketPVP_01");
					else
						icon = ("Interface\\ICONS\\INV_Jewelry_TrinketPVP_02");
					end
				end
				
				historyIcon.texture:SetTexture(icon);
				historyIcon:SetAlpha(0);
				ArenaLive_CastHistoryIcon_Update(historyIcon, castHistory);
				historyIcon:Show();	
				
				if ( castType == "hardCast" ) then
					historyIcon.isCasting = true;
					historyIcon.lineID = lineID;
					historyIcon.castingAnimationGroup:Play();
				elseif ( castType == "channel" ) then
					historyIcon.isChanneling = true;
					historyIcon.castingAnimationGroup:Play();
				else
					-- Instant
					historyIcon.fadeInAnimationGroup:Play();
				end
				
				historyIcon.isActive = true;
				iconFound = true;
				castHistory.newestIcon = historyIcon;
			end
		
		else

			if (not dontMove ) then
			
				historyIcon.timesMoved = historyIcon.timesMoved + 1;

				if ( historyIcon.moveAnimationGroup:IsPlaying() ) then
					-- If the icon is already moving we just add the default offset to the current transitionway
					local xOffset, yOffset = historyIcon.moveAnimationGroup.translation:GetOffset();
					xOffset = xOffset + historyIcon.defaultxOffset;
					historyIcon.moveAnimationGroup.translation:SetOffset(xOffset/1.35, yOffset);
				else
					historyIcon.moveAnimationGroup:Play();
				end
				
				
				if (not historyIcon.isFading and historyIcon.timesMoved > ARENALIVE_DATABASE[castHistory.addonName]["frameTypes"][castHistory.frameType]["castHistory"]["shownIcons"] - 1 ) then
					historyIcon.lastMove = true;
					historyIcon.fadeOutAnimationGroup:Play();
				end
			
			end		
		
		end
	end

	-- If we couldn't get a free Icon, we simply will create an extra one.
	if ( not iconFound) then
		
		castHistory.numIcons = castHistory.numIcons + 1;
		
		local iconName = "Icon"..castHistory.numIcons;
		local iconShortCut = "icon"..castHistory.numIcons;
		
		historyIcon = ArenaLive_CastHistoryIcon_Create(castHistory, iconName, iconShortCut, castHistory.moveDirection, castHistory.iconTemplate);
		
		historyIcon.isActive = true;
		historyIcon.spellID = spellID;
		historyIcon.texture:SetTexture(icon);
		historyIcon:SetAlpha(0);
		historyIcon:Show();	
				
		if ( castType == "hardCast" ) then
			historyIcon.isCasting = true;
			historyIcon.lineID = lineID;
			historyIcon.castingAnimationGroup:Play();
		elseif ( castType == "channel" ) then
			historyIcon.isChanneling = true;
			historyIcon.castingAnimationGroup:Play();
		else
			-- Instant
			historyIcon.fadeInAnimationGroup:Play();
		end
			
	end	
	
end

function ArenaLive_CastHistory_UpdateAllIcons (self)
	
	-- First reset the castbar, so no icons are shown.
	ArenaLive_CastHistory_ResetAllIcons(self)
	
	for i=1, self.numIcons do
		ArenaLive_CastHistoryIcon_Update(self["icon"..i], self)
	end

end

function ArenaLive_CastHistory_ResetAllIcons (self)

	for i=1, self.numIcons do
		ArenaLive_CastHistoryIcon_Reset(self["icon"..i], self)
	end

end



function ArenaLive_CastHistoryIcon_Create (castHistory, iconName, iconShortCut, moveDirection, iconTemplate)
	
	local iconSize = ARENALIVE_DATABASE[castHistory.addonName]["frameTypes"][castHistory.frameType]["castHistory"]["iconSize"];
	local xOffset = ArenaLive_CastHistoryIcon_GetDefaultOffset(iconSize, moveDirection);
	local prefix = castHistory:GetName();
	
	ArenaLive_SystemMessage("Create new castHistory icon: "..castHistory:GetName(), true);
	CreateFrame("Button", prefix..iconName, castHistory, iconTemplate);
	castHistory[iconShortCut] = _G[prefix..iconName];
	castHistory[iconShortCut].parent = castHistory;
	
	if ( castHistory[iconShortCut] ) then
		
		
		castHistory[iconShortCut]:SetSize(iconSize, iconSize);
		
		if ( moveDirection == "left" ) then
			castHistory[iconShortCut]:SetPoint("RIGHT");
		else
			castHistory[iconShortCut]:SetPoint("LEFT");
		end

		castHistory[iconShortCut].texture = _G[prefix..iconName.."Texture"];
		castHistory[iconShortCut].border = _G[prefix..iconName.."Border"];
		castHistory[iconShortCut].lockOutTexture = _G[prefix..iconName.."LockOutTexture"];
		castHistory[iconShortCut].castingAnimationGroup = _G[prefix..iconName.."Casting"];
		castHistory[iconShortCut].moveAnimationGroup = _G[prefix..iconName.."Move"];
		castHistory[iconShortCut].moveAnimationGroup.translation = _G[prefix..iconName.."MoveTranslation"];
		castHistory[iconShortCut].fadeInAnimationGroup = _G[prefix..iconName.."FadeIn"];
		castHistory[iconShortCut].fadeOutAnimationGroup = _G[prefix..iconName.."FadeOut"];
		
		castHistory[iconShortCut].border:SetSize(iconSize+1, iconSize+1);

		castHistory[iconShortCut].moveAnimationGroup.translation:SetOffset(xOffset/1.35, 0);
		
		castHistory[iconShortCut].timesMoved = 0;
		castHistory[iconShortCut].defaultxOffset = xOffset;
		castHistory[iconShortCut].duration = ARENALIVE_DATABASE["options"]["castHistoryIconDuration"];
		
		castHistory[iconShortCut]:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
		castHistory[iconShortCut]:RegisterEvent("UNIT_SPELLCAST_STOP");
		castHistory[iconShortCut]:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP");
		castHistory[iconShortCut]:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
		castHistory[iconShortCut]:SetScript("OnEvent", ArenaLive_CastHistoryIcon_OnEvent);
		
		return castHistory[iconShortCut];
	end
end

function ArenaLive_CastHistoryIcon_GetDefaultOffset(iconSize, moveDirection)
	
	if ( not moveDirection or not iconSize ) then
		return;
	end
	
	local xOffset;
	
	if ( moveDirection == "left" ) then
		xOffset = -iconSize - defaultxOffsetModifier;
	else
		xOffset = iconSize + defaultxOffsetModifier;
	end
	
	return xOffset;
end

function ArenaLive_CastHistoryIcon_OnEvent (self, event, ...)
		
	if (not self.isActive ) then
		return;
	end
	local unit, spell, rank, lineID, spellID;
	
	
	if ( event ~= "COMBAT_LOG_EVENT_UNFILTERED" ) then
		unit, spell, rank, lineID, spellID = ...;
	end
	
	if ( self.isCasting and ( event == "UNIT_SPELLCAST_SUCCEEDED" ) ) and ( select(4, ...) == self.lineID ) and ( select(1, ...) == self.parent.unit ) then
		self.isCasting = nil;
		self.lineID = nil;
		self.isAbandoned = nil;
		self.castingAnimationGroup:Finish();
		self:SetScript("OnUpdate", ArenaLive_CastHistoryIcon_OnUpdate);
	elseif ( self.isCasting and ( event == "UNIT_SPELLCAST_STOP" ) and ( select(4, ...) == self.lineID ) and ( select(1, ...) == self.parent.unit ) ) then
		
		self.isCasting = nil;
		self.lineID = nil;
		self.isAbandoned = true;
		self.castingAnimationGroup:Finish();
		
	elseif ( self.isChanneling and event == "UNIT_SPELLCAST_CHANNEL_STOP" ) and ( select(1, ...) == self.parent.unit ) then
	
		self.isChanneling = nil;
		self.isAbandoned = nil;
		self.castingAnimationGroup:Finish();
		self:SetScript("OnUpdate", ArenaLive_CastHistoryIcon_OnUpdate);
		
	elseif ( event == "COMBAT_LOG_EVENT_UNFILTERED" ) and ( self.spellID ) then
		
		local event2 = select(2, ...);
		local sourceGUID = select(4, ...);
		local destGUID = select(8, ...);
		local spellID = select(12, ...);
		local extraSpellID = select(15, ...);
		local class, classFilename, race, raceFilename, sex, name, realm;
		
		if ( destGUID ~= "0x0000000000000000" and destGUID ~= "" and destGUID ) then
			class, classFilename, race, raceFilename, sex, name, realm = GetPlayerInfoByGUID(destGUID)
		end		
		
		if ( ( event2 == "SPELL_DAMAGE" or event2 == "SPELL_HEAL" or event2 == "SPELL_CAST_SUCCESS" ) and sourceGUID == self.parent.guid and spellID == self.spellID ) then
			
			if ( RAID_CLASS_COLORS[classFilename] ) then
				self.border:Show();
				self.border:SetVertexColor(RAID_CLASS_COLORS[classFilename].r, RAID_CLASS_COLORS[classFilename].g, RAID_CLASS_COLORS[classFilename].b, 1);
			else
				self.border:Hide();
			end
			
			self.spellID = nil;
			
		elseif( event2 == "SPELL_INTERRUPT" and destGUID == self.parent.guid and extraSpellID == self.spellID ) then
			
			self.isAbandoned = nil;
			self.border:Show();
			self.lockOutTexture:Show();
			self.border:SetVertexColor(1, 0, 0, 1);
			self.spellID = nil;		
		
		end
	
	end
end



function ArenaLive_CastHistoryIcon_OnUpdate (self, elapsed)

	if ( self.isActive and not self.isFading ) then
		self.duration = self.duration - elapsed;
		
		if ( self.duration <= 0 ) then
			-- Time's up, so we fade the button out.
			self:SetScript("OnUpdate", nil);
			self.isFading = true;
			
			if ( not self.lastMove ) then
				self.fadeOutAnimationGroup:Play();
			end
		end
	end

end

function ArenaLive_CastHistoryIcon_Update(icon, castHistory)
	local moveDirection = castHistory.moveDirection;
	local iconSize = ARENALIVE_DATABASE[castHistory.addonName]["frameTypes"][castHistory.frameType]["castHistory"]["iconSize"];
	local defaultxOffset = ArenaLive_CastHistoryIcon_GetDefaultOffset(iconSize, moveDirection);
	
	icon:SetSize(iconSize, iconSize);
	icon.border:SetSize(iconSize+1, iconSize+1);
	icon.defaultxOffset = defaultxOffset;
	icon.moveAnimationGroup.translation:SetOffset(defaultxOffset/1.35, 0);
end

function ArenaLive_CastHistoryIcon_Reset (self, castHistory)
	
	if ( not self ) or ( not castHistory ) then
		return;
	end
	
	self.moveAnimationGroup:Stop();
	self.fadeInAnimationGroup:Stop();
	self.fadeOutAnimationGroup:Stop();
	
	self:Hide();
	self:SetScript("OnUpdate", nil);
	self.texture:SetTexture();
	self.border:Hide();
	self.lockOutTexture:Hide();
	self.moveAnimationGroup.translation:SetOffset(self.defaultxOffset/1.35, 0);
	self:SetAlpha(1);
	
	self:ClearAllPoints();
	
	if ( castHistory.moveDirection == "left" ) then
		self:SetPoint("RIGHT");
	else
		self:SetPoint("LEFT");
	end
	
	self.timesMoved = 0;
	self.duration = ARENALIVE_DATABASE["options"]["castHistoryIconDuration"];
	self.isFading = nil;
	self.isAbandoned = nil;
	self.lastMove = nil;
	self.isCasting = nil;
	self.isChanneling = nil;
	self.spellID = nil;
	self.lineID = nil;
	self.isActive = nil;
end



function ArenaLive_CastHistoryIcon_AnimationHandler(self, parent, animationType)

	if ( animationType == "FadeIn" ) then
		parent:SetAlpha(1);
		parent:SetScript("OnUpdate", ArenaLive_CastHistoryIcon_OnUpdate);
		
	elseif ( animationType == "FadeOut" ) then
			parent:Hide();
			parent:SetAlpha(0);
		
		if ( not parent.lastMove and parent.isFading and parent.isActive ) then
			ArenaLive_CastHistoryIcon_Reset(parent, parent.parent);
		end
		
	elseif ( animationType == "Move" ) then
		-- Reset animation's xOffset to default
		local point, relativeTo, relativePoint, xOffset, yOffset = parent:GetPoint();
		local animationxOffset, animationyOffset = self.translation:GetOffset();
		xOffset = parent.defaultxOffset * parent.timesMoved;
		yOffset = yOffset + animationyOffset;
		parent:ClearAllPoints();
		parent:SetPoint(point, relativeTo, relativePoint, xOffset, yOffset);
		self.translation:SetOffset(parent.defaultxOffset/1.35, 0);
		if ( parent.lastMove and parent.isActive ) then
			ArenaLive_CastHistoryIcon_Reset(parent, parent.parent);
		end
	elseif ( animationType == "Casting" ) then
		
		if ( not parent.isAbandoned ) then
		
			parent.fadeInAnimationGroup:Play();
		
		else
			
			ArenaLive_CastHistoryIcon_Reset(parent, parent.parent);
		
		end
		
	end

end